local trackedprefabs = { "minotaur" }
local trackedprefabs_lookup = {}
for _, v in ipairs(trackedprefabs) do
	trackedprefabs_lookup[v] = true
end

local function GetEntityDebugString(inst)
	assert(inst ~= nil)
	local prefab = inst.prefab
	local pos = inst:GetPosition()

	return string.format("%s (%s,%s,%s)", tostring(prefab), tostring(pos.x), tostring(pos.y), tostring(pos.z))
end


local ResourceRespawnerComponent = Class(function(self, inst)
	self.inst = inst
	self.isdebug = 1 -- modmain sets the actual value
	self.spawnFreeZone = 4 -- modmain sets the actual value
	self.doubleRenewDelay = 0 -- modmain sets the actual value via SetRegenSpeed()
	self.respawnDay = {}
	self.currentDay = 0 -- zero-based indexing for easier modulo math
	self.currentHour = 0 -- zero-based indexing for easier modulo math
	self.shouldAdd = true

	self.restoreTurfs = false -- modmain sets the actual value
	self.turfs = nil
	self.turfsRestoreDay = {}
	self.turfsRestoreDelay = 0 -- modmain sets the actual value via SetRegenSpeed()

	--Won't be exactly in sync with the clock (e.g. after saving/loading), but we don't care and there are enough events happening on clock cycle change already
	self.inst:DoPeriodicTask(TUNING.SEG_TIME, function() self:ClockChanged() end)

	-- schedule turfs initialization
	-- SharedIslands mod does its initialization at 1s, wait till after, just to be safe
	self.inst:DoTaskInTime(2, function() self:InitTurfs() end)
end)

-- set doubleRenewDelay and turfsRestoreDelay
-- initializes respawnDay and turfsRestoreDay tables (with empty tables)
-- @param speed number
function ResourceRespawnerComponent:SetRegenSpeed(speed)
	assert(type(speed) == "number", string.format("Parameter speed '%s' has type %s, expected number.", tostring(speed), type(speed)))
	-- check if the respawnDay table is effectively empty
	if self.respawnDay ~= nil then
		assert(type(self.respawnDay) == "table", string.format("self.respawnDay '%s' has type %s, expected table.", tostring(self.respawnDay), type(self.respawnDay)))
		for i, t in pairs(self.respawnDay) do
			assert(type(t) == "table", string.format("self.respawnDay[%s] '%s' has type %s, expected table", tostring(i), tostring(t), type(t)))
			assert(next(t) == nil, string.format("self.respawnDay[%s] is not empty - %s", tostring(i), tabletodictstring(t)))
		end
	end

	-- check if the turfsRestoreDay table is effectively empty
	if self.turfsRestoreDay ~= nil then
		assert(type(self.turfsRestoreDay) == "table", string.format("self.turfsRestoreDay '%s' has type %s, expected table.", tostring(self.turfsRestoreDay), type(self.turfsRestoreDay)))
		for i, t in pairs(self.turfsRestoreDay) do
			assert(type(t) == "table", string.format("self.turfsRestoreDay[%s] '%s' has type %s, expected table", tostring(i), tostring(t), type(t)))
			assert(next(t) == nil, string.format("self.turfsRestoreDay[%s] is not empty - %s", tostring(i), tabletodictstring(t)))
		end
	end

	self.doubleRenewDelay = speed * 2
	self.turfsRestoreDelay = speed * 2 - 1 -- speed*2 would mean today, -1 makes it almost the whole cycle
	self.respawnDay = {}
	self.turfsRestoreDay = {}
	for i = 1, self.doubleRenewDelay do
		table.insert(self.respawnDay, {})
		table.insert(self.turfsRestoreDay, {})
	end
end

function ResourceRespawnerComponent:__tostring()
	return string.format("ResourceRespawnerComponent - inst=%s, isdebug=%s, spawnFreeZone=%s, doubleRenewDelay=%s, currentDay=%s, currentHour=%s, shouldAdd=%s, restoreTurfs=%s, turfsRestoreDelay=%s",
		tostring(self.inst), tostring(self.isdebug), tostring(self.spawnFreeZone), tostring(self.doubleRenewDelay), tostring(self.currentDay), tostring(self.currentHour), tostring(self.shouldAdd), tostring(self.restoreTurfs), tostring(self.turfsRestoreDelay)
	)
end

function ResourceRespawnerComponent:AddInstance(inst, respawndelay)
	local x, y, z = inst.Transform:GetWorldPosition()
	if x ~= 0 and x ~= nil then -- Items added with SpawnPrefab that have no location in the world should not be regenerated
		local targetDay = (self.currentDay + respawndelay) % self.doubleRenewDelay + 1 -- currentDay is zero-based, respawnDay is one-based
		table.insert(self.respawnDay[targetDay], { inst.prefab, x, y, z })
	elseif trackedprefabs_lookup[inst.prefab] then
		print(string.format("[WorldResourceRegeneration] Tracked prefab %s not recheduled at coordinates (%s,%s,%s).",
			tostring(inst.prefab), tostring(x), tostring(y), tostring(z)
		))
	end
end

function ResourceRespawnerComponent:SpawnInstance(inst)
	local ents = TheSim:FindEntities(inst[2], inst[3], inst[4], self.spawnFreeZone, { "structure" })  -- 1=inst.prefab, 2=x, 3=y, 4=z
	for i, v in ipairs(ents) do -- Don't respawn in the middle of a base
		if trackedprefabs_lookup[inst[1]] then
			local entitydbgstr = GetEntityDebugString(v)
			print(string.format("[WorldResourceRegeneration] Tracked prefab %s not respawned at coordinates (%s,%s,%s) due to %s.",
				tostring(inst[1]), tostring(inst[2]), tostring(inst[3]), tostring(inst[4]), entitydbgstr
			))
		end
		return false
	end
	local newPrefab = SpawnPrefab(inst[1])
	if newPrefab then
		newPrefab.Transform:SetPosition(inst[2], inst[3], inst[4])
	elseif trackedprefabs_lookup[inst[1]] then
		print(string.format("[WorldResourceRegeneration] Tracked prefab %s failed to respawn at coordinates (%s,%s,%s).",
			tostring(inst[1]), tostring(inst[2]), tostring(inst[3]), tostring(inst[4])
		))
	end
	return true
end

-- tries to add turf at given coordinates to the respawn table
-- @param x number  x-coordinate of the turf (zero-based index)
-- @param u number  y-coordinate of the turf (zero-based index)
function ResourceRespawnerComponent:AddTurfRegen(x, y)
	assert(type(x) == "number", string.format("Parameter x '%s' has type %s, expected number.", tostring(x), type(x)))
	assert(type(y) == "number", string.format("Parameter y '%s' has type %s, expected number.", tostring(y), type(y)))
	if not self.restoreTurfs then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Trying to add a turf restoration, but self.restoreTurfs=%s", tostring(x), tostring(y), tostring(self.restoreTurfs)))
		end
		return
	end

	if type(self.turfs) ~= "table" then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Trying to add a turf restoration, but self.turfs=%s", tostring(x), tostring(y), tostring(self.turfs)))
		end
		return
	end

	-- shift to one-based indexing
	local x1 = x + 1
	local y1 = y + 1

	-- check if there is an entry for given coordinates in the turfs table
	-- if none is found, it means there was a dirt previously (or possibly an ocean?)
	local tblx = self.turfs[x1]
	if tblx == nil then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Trying to add a turf coordinates, but row %s was not found in the turfs table.",
				tostring(x), tostring(y), tostring(x1)
			))
		end
		return
	end
	local tile = tblx[y1]
	if tile == nil then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Trying to add a turf coordinates, but column %s was not found in the turfs[%s] table.",
				tostring(x), tostring(y), tostring(y1), tostring(x1)
			))
		end
		return
	end

	-- cancel previously scheduled turf at the same coordinates
	for i, day in ipairs(self.turfsRestoreDay) do
		for j = #day, 1, -1 do
			local itemx, itemy = unpack(day[j])
			if itemx == x and itemy == y then
				if self.isdebug > 0 then
					print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Cancelling previously added turf (day %s, index %s).",
						tostring(x), tostring(y), tostring(i), tostring(j)
					))
				end
				table.remove(day, j)
			end
		end
	end

	local targetDay = (self.currentDay + self.turfsRestoreDelay) % self.doubleRenewDelay + 1 -- currentDay is zero-based, turfsRestoreDay is one-based
	table.insert(self.turfsRestoreDay[targetDay], {x, y})
	if self.isdebug > 0 then
		print(string.format("[ResourceRespawnerComponent:AddTurfRegen(%s,%s)] Turf coordinates added, targetDay=%s.", tostring(x), tostring(y), tostring(targetDay)))
	end
end

-- try to restore turf to the saved one
-- @param x number  x-coordinate of the turf (zero-based index)
-- @param u number  y-coordinate of the turf (zero-based index)
function ResourceRespawnerComponent:RestoreTurf(x, y)
	assert(type(x) == "number", string.format("Parameter x '%s' has type %s, expected number.", tostring(x), type(x)))
	assert(type(y) == "number", string.format("Parameter y '%s' has type %s, expected number.", tostring(y), type(y)))
	if not self.restoreTurfs then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Trying to restore a turf, but self.restoreTurfs=%s", tostring(x), tostring(y), tostring(self.restoreTurfs)))
		end
		return
	end

	if type(self.turfs) ~= "table" then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Trying to restore a turf, but self.turfs=%s", tostring(x), tostring(y), tostring(self.turfs)))
		end
		return
	end

	-- shift to one-based indexing
	local x1 = x + 1
	local y1 = y + 1

	-- check if there is an entry for given coordinates in the turfs table
	-- if none is found, it means there was a dirt previously
	local tblx = self.turfs[x1]
	if tblx == nil then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Trying to restore a turf, but row %s was not found in the turfs table.",
				tostring(x), tostring(y), tostring(x1)
			))
		end
		return
	end

	local tile = tblx[y1]
	if tile == nil then
		if self.isdebug > 0 then
			print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Trying to restore a turf, but column %s was not found in the turfs[%s] table.",
				tostring(x), tostring(y), tostring(y1), tostring(x1)
			))
		end
		return
	end

	local tilename = table.reverselookup(GROUND, tile)
	if tilename == nil then
		print(string.format("[WorldResourceRespawner] Trying to restore turf at coordinates [%s,%s], original tile id is %s, but it's not found in the GROUND table.",
			tostring(x), tostring(y), tostring(tile)
		))
	end

	local world = TheWorld
	local map = world.Map
	local minimap = world.minimap.MiniMap

	local original_tile_type = map:GetTile(x, y)

	if original_tile_type ~= GROUND.DIRT then
		if self.isdebug > 0 then
			local original_tilename = table.reverselookup(GROUND, original_tile_type)
			print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Current tile type is %s (%s), turf restoration skipped.",
				tostring(x), tostring(y), tostring(original_tile_type), tostring(original_tilename)
			))
		end
		return
	end

	map:SetTile(x, y, tile)
	map:RebuildLayer(original_tile_type, x, y)
	map:RebuildLayer(tile, x, y)

	minimap:RebuildLayer(original_tile_type, x, y)
	minimap:RebuildLayer(tile, x, y)

	if self.isdebug > 0 then
		print(string.format("[ResourceRespawnerComponent:RestoreTurf(%s,%s)] Turf %s (%s) restored.",
			tostring(x), tostring(y), tostring(tile), tostring(tilename)
		))
	end
end

function ResourceRespawnerComponent:ClockChanged()
	self.shouldAdd = false -- World is created, all entities that should be added are now added, so we are done adding new ones

	self.currentHour = self.currentHour + 1
	if self.currentHour >= 16 then -- 16 segments per day
		self.currentHour = 0
		self.currentDay = self.currentDay + 1
		if self.currentDay >= self.doubleRenewDelay then
			self.currentDay = 0
		end
	end

	local respawnToday = self.respawnDay[self.currentDay + 1] -- currentDay is zero-based, respawnDay is one-based
	local numEntitiesToSpawn = #respawnToday / (16 - self.currentHour) -- respawn 1/16th of the entities for this day and slowly increase this to 100%
	self.shouldAdd = true -- Add the respawned ones to the list again
	for i = 1, numEntitiesToSpawn do
		self:SpawnInstance(respawnToday[1])
		table.remove(respawnToday, 1)
	end
	self.shouldAdd = false -- Prevent other prefabs being created to also get added to the list

	-- turfs restoration
	if self.restoreTurfs then
		local turfsRestoreToday = self.turfsRestoreDay[self.currentDay + 1] -- currentDay is zero-based, turfsRestoreDay is one-based
		local numTurfsToRestore = #turfsRestoreToday / (16 - self.currentHour) -- restore 1/16th of the turfs for this day and slowly increase this to 100%
		for i = 1, numTurfsToRestore do
			self:RestoreTurf(unpack(turfsRestoreToday[1]))
			table.remove(turfsRestoreToday, 1)
		end
	end
end

function ResourceRespawnerComponent:OnSave()
	local data = {
		shouldAdd = self.shouldAdd,
		currentHour = self.currentHour,
		currentDay = self.currentDay,
		respawnDay = self.respawnDay
	}

	self:tprint(self.respawnDay, 1)
	if self.restoreTurfs then
		data.turfs = self.turfs
		data.turfsRestoreDay = self.turfsRestoreDay
		self:tprint(self.turfsRestoreDay, 1)
	end

	return data
end

function ResourceRespawnerComponent:OnLoad(data)
	if data then
		self.currentHour = data.currentHour
		self.currentDay = data.currentDay

		-- Fixes a crash when regeneration interval is lowered too much (less than current day/2). Fix by Muche.
		if self.currentDay >= self.doubleRenewDelay then
			self.currentDay = 0
		end

		self.respawnDay = data.respawnDay
		self.shouldAdd = data.shouldAdd

		local daysToRemove = #self.respawnDay - self.doubleRenewDelay
		for i = 1, daysToRemove do -- In case renew delay has been reduced move all items awaiting spawn to new spawn date
			for j = 1, #self.respawnDay[self.doubleRenewDelay + 1] do
				table.insert(self.respawnDay[self.doubleRenewDelay], self.respawnDay[self.doubleRenewDelay + 1][j])
			end
			table.remove(self.respawnDay, self.doubleRenewDelay + 1)
		end

		local daysToAdd = self.doubleRenewDelay - #self.respawnDay
		for i = 1, daysToAdd do  -- In case renew delay has been increased make table big enough
			table.insert(self.respawnDay, {})
		end

		if self.restoreTurfs then
			if data.turfs ~= nil then
				self:CheckTurfs(data.turfs)
				self.turfs = data.turfs
			end
			if data.turfsRestoreDay ~= nil then
				self.turfsRestoreDay = data.turfsRestoreDay
			end

			local daysToRemove = #self.turfsRestoreDay - self.doubleRenewDelay
			for i = 1, daysToRemove do -- In case renew delay has been reduced move all items awaiting spawn to new spawn date
				for j = 1, #self.turfsRestoreDay[self.doubleRenewDelay + 1] do
					table.insert(self.turfsRestoreDay[self.doubleRenewDelay], self.turfsRestoreDay[self.doubleRenewDelay + 1][j])
				end
				table.remove(self.turfsRestoreDay, self.doubleRenewDelay + 1)
			end

			local daysToAdd = self.doubleRenewDelay - #self.turfsRestoreDay
			for i = 1, daysToAdd do  -- In case renew delay has been increased make table big enough
				table.insert(self.turfsRestoreDay, {})
			end
		end
	end
	self:tprint(self.respawnDay, 1)
	if self.restoreTurfs then
		self:tprint(self.turfsRestoreDay, 1)
	end
end

-- Analyse table structure
function ResourceRespawnerComponent:tprint(tbl, indent)
	for k, v in pairs(tbl) do
		local formatting = string.rep("  ", indent) .. k .. ": "
		if type(v) == "table" then
			if self.isdebug > 0 then
				print(formatting)
			end
			self:tprint(v, indent+1)
		else
			if self.isdebug > 0 then
				print(formatting .. v)
			end
		end
	end
end

-- initialize turfs
-- if you want to reinitialize turfs (i.e. make current world's turfs the original ones):
--   disable the turfs regeneration option, load the world, save the world, and enable the option
--   or run in the debug console: TheWorld.components.ResourceRespawnerComponent.turfs=nil; TheWorld.components.ResourceRespawnerComponent:InitTurfs()
function ResourceRespawnerComponent:InitTurfs()
	if self.isdebug > 0 then
		print(string.format("[ResourceRespawnerComponent:InitTurfs()] restoreTurfs=%s, turfs=%s",
			tostring(self.restoreTurfs), tostring(self.turfs)
		))
	end

	if not self.restoreTurfs or self.turfs ~= nil then
		-- turfs regeneration is disabled or turfs have already been loaded
		return
	end

	assert(TheWorld ~= nil and TheWorld.Map ~= nil, "TheWorld.Map not found.")
	local GROUND = GROUND
	assert(GROUND.DIRT ~= nil, "Core tile type GROUND.DIRT not found.")
	assert(GROUND.IMPASSABLE ~= nil, "Core tile type GROUND.IMPASSABLE not found.")
	assert(GROUND.INVALID ~= nil, "Core tile type GROUND.INVALID not found.")

	local turfs = {}
	local map = TheWorld.Map
	local maxx, maxy = map:GetSize()
	for x = 1, maxx do
		local tblx = {}
		for y = 1, maxy do
			local tile = map:GetTile(x - 1, y - 1) -- GetTile uses zero-based indexes
			-- skip non-terraformable tile types
			if tile ~= GROUND.DIRT and tile ~= GROUND.IMPASSABLE and tile ~= GROUND.INVALID then
				tblx[y] = tile
			end
		end

		-- skip empty table
		if next(tblx) then
			turfs[x] = tblx
		end
	end

	self.turfs = turfs
end

-- check validity of turfs
function ResourceRespawnerComponent:CheckTurfs(turfs)
	assert(TheWorld ~= nil and TheWorld.Map ~= nil, "TheWorld.Map not found.")
	assert(type(turfs) == "table", string.format("Parameter turfs '%s' has type %s, expected table.", tostring(turfs), type(turfs)))
	local GROUND_LOOKUP = table.invert(GROUND)
	local maxx, maxy = TheWorld.Map:GetSize()
	local unknowntiles = {} -- table of tile ids that are in the turfs table but not in the current GROUND table

	for x, tblx in pairs(turfs) do
		assert(type(x) == "number", string.format("turfs index '%s' has type %s, expected number", tostring(x), type(x)))
		assert(x > 0 and x <= maxx, string.format("turfs index '%s' is out of limits [0,%s]", tostring(x), type(maxx)))
		assert(type(tblx) == "table", string.format("turfs[%s] '%s' has type %s, expected table", tostring(x), tostring(tblx), type(tblx)))
		for y, tile in pairs(tblx) do
			assert(type(y) == "number", string.format("turfs[%s] index '%s' has type %s, expected number", tostring(x), tostring(y), type(y)))
			assert(y > 0 and y <= maxy, string.format("turfs[%s] index '%s' is out of limits [0,%s]", tostring(x), tostring(y), type(maxy)))
			assert(type(tile) == "number", string.format("turfs[%s][%s] '%s' has type %s, expected number", tostring(x), tostring(y), tostring(tile), type(tile)))
			if GROUND_LOOKUP[tile] == nil then
				-- tile is unknown
				unknowntiles[tile] = true
			end
		end
	end

	if next(unknowntiles) ~= nil then
		local s = nil
		for tile, _ in pairs(unknowntiles) do
			if s == nil then
				s = tostring(tile)
			else
				s = s .. "," .. tostring(tile)
			end
		end
		print("[ResourceRespawnerComponent:CheckTurfs()] turfs table has unknown tiles (not found in GROUND table): " .. s)
	end
end

return ResourceRespawnerComponent
